/*
The copyright in this software is being made available under the BSD License, included below. This software may be subject to other third party and contributor rights, including patent rights, and no such rights are granted under this license.
 
Copyright (c) 2013, ITU and ISO/IEC
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
  * Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
  * Neither the name of the ITU and ISO/IEC nor the names of its
    contributors may be used to endorse or promote products derived from this
    software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include "yuv_proc.h"
#include <string.h>
#include <math.h>
#pragma warning( disable: 4996 )

inline byte ClipToByte( int iX ) {return (byte)( iX > 0 ? ( iX < 255? iX : 255 ) : 0 ); }


void ComputePSNRforYUV(byte *pYUV, byte *pRefYUV, int srcW, int srcH, YUVFormat frameFormat, double *psnrY, double *psnrU, double *psnrV)
{
    long ssdY=0,ssdU=0,ssdV=0;
    int x,y;
    int frameSize = srcW*srcH;
    byte *pU, *pV, *pRefU,*pRefV;
    int chromaShift = 0;

    if(frameFormat == YUV_444){
        pU = pYUV + frameSize;
        pV = pU + frameSize;
        pRefU = pRefYUV + frameSize;
        pRefV = pRefU + frameSize;
    }
    else if(frameFormat == YUV_420){
        pU = pYUV + frameSize;
        pV = pU + frameSize/4;
        chromaShift = 1;
        pRefU = pRefYUV + frameSize;
        pRefV = pRefU + frameSize/4;
    }
    else {
        printf("PSNR only supported for 444 and 420 YUVs\n");
        return;
    }

    //PSNR for Y
    for( y = 0; y < srcH; y++ ) {
        for( x = 0; x < srcW; x++ ) {
            int diff = (int)( pYUV[x] - pRefYUV[x] );
            ssdY += diff * diff;
        }
        pYUV    += srcW;
        pRefYUV += srcW;
    }

    //PSNR for U
    for( y = 0; y < (srcH>>chromaShift); y++ ) {
        for( x = 0; x < (srcW>>chromaShift); x++ ) {
            int diff = (int)( pU[x] - pRefU[x] );
            ssdU += diff * diff;
        }
        pU    += (srcW>>chromaShift);
        pRefU += (srcW>>chromaShift);
    }

    //PSNR for V
    for( y = 0; y < (srcH>>chromaShift); y++ ) {
        for( x = 0; x < (srcW>>chromaShift); x++ ) {
            int diff = (int)( pV[x] - pRefV[x] );
            ssdV += diff * diff;
        }
        pV    += (srcW>>chromaShift);
        pRefV += (srcW>>chromaShift);
    }

  int maxval = 255;
  double fmaxValY = (double) maxval * maxval * frameSize;
  double fmaxValC = (double) maxval * maxval * frameSize / ((1<<chromaShift)*(1<<chromaShift));
  *psnrY = ( ssdY ? 10*log10( fmaxValY / (double)ssdY ) : 100 );
  *psnrU = ( ssdU ? 10*log10( fmaxValC / (double)ssdU ) : 100 );
  *psnrV = ( ssdV ? 10*log10( fmaxValC / (double)ssdV ) : 100 );

}


byte Round( double dIn )
{
    int iOut = (int)( dIn + 0.5 );
    return (byte)((iOut < 0) ? 0 : ((iOut > 255) ? 255 : iOut));
}

//NOTE: Only works with YUV/RGB that have even number of samples every row
void ColorConversionRGBtoYUV444( byte *pRGB,byte *pYuv, int srcH, int srcW,int iDstStride)
{
#ifdef BT601
    double Kr = 0.299;
    double Kb = 0.114;
#else
    double Kr = 0.2126;
    double Kb = 0.0722;
#endif
    double Kg = 1.0 - Kr - Kb;
    double Dcb = ( 2.0 *  ( 1.0 - Kb ) );
    double Dcr = ( 2.0 *  ( 1.0 - Kr ) );

    int x,y;
    byte *pDestY = pYuv;
    byte *pDestU = pYuv + srcH*srcW;
    byte *pDestV = pYuv + srcH*srcW*2;


    for ( y = 0; y < srcH; ++y ) {
        for ( x = 0; x < srcW; ++x ) 
        {
            int B =  pRGB[4*x + 0];
            int G =  pRGB[4*x + 1];
            int R =  pRGB[4*x + 2];

            double Ey  = ( Kr * R + Kg * G + Kb * B ) / 255.0;
            double Epb = ( B / 255.0 - Ey ) / Dcb;
            double Epr = ( R / 255.0 - Ey ) / Dcr;            

            *pDestY++ = Round( 219.0 * Ey + 16.0 );
            *pDestU++ = Round( 224.0 * Epb + 128.0 );
            *pDestV++ = Round( 224.0 * Epr + 128.0 );
        }
        pRGB   += iDstStride*4;
        pDestY += (iDstStride-srcW);
        pDestU += (iDstStride-srcW);
        pDestV += (iDstStride-srcW);
    }

}

//NOTE: Only works with YUV/RGB that have even number of samples every row
void ColorConversionYUV444toRGB32(byte *pY, byte *pU,byte *pV,byte *pRGB,int srcH, int srcW,int iDstStride)
{
    int x,y;
#ifdef BT601
    double Kr = 0.299;
    double Kb = 0.114;
#else
    double Kr = 0.2126;
    double Kb = 0.0722;
#endif
    double Kg = 1.0 - Kr - Kb;
    double Dcb = ( 2.0 *  ( 1.0 - Kb ) );
    double Dcr = ( 2.0 *  ( 1.0 - Kr ) );


    for ( y = 0; y < srcH; ++y ) {
        for ( x = 0; x < srcW; ++x ) 
        {
            double Ey  = ( *pY++ - 16.0 ) / 219.0;
            double Epb = ( *pU++ - 128.0 ) / 224.0;
            double Epr = ( *pV++ - 128.0 ) / 224.0;

            double Fr = 255.0 * ( Dcr * Epr + Ey );
            double Fb = 255.0 * ( Dcb * Epb + Ey );
            double Fg = ( ( 255.0 * Ey ) - Fr * Kr - Kb * Fb ) / Kg;

            *(pRGB+ 4*x)    = Round( Fr );
            *(pRGB+ 4*x+1)  = Round( Fg );
            *(pRGB+ 4*x+2)  = Round( Fb );    
        }
        pRGB   += iDstStride*4;
        pY += (iDstStride-srcW);
        pU += (iDstStride-srcW);
        pV += (iDstStride-srcW);
    }
}

//------------------------------------------------------------------
// FRAME PACKING MODULES
//------------------------------------------------------------------
void FramePackDirectMode(byte *pSrc, byte *pDstY, byte *pDstU, byte *pDstV, int srcH, int srcW, 
                         int iSrcStrideUV, int iDstStride,int iDstStrideUV, double dAlpha, double dBeta, bool bAntiAlias)
{
    int x,y;
    byte *pSrcInp       = pSrc;
    byte *pDestY420		= pDstY;
    byte *pDestU420		= pDstU;
    byte *pDestV420		= pDstV;
    byte *pDestHalfU    = pDstY+iDstStride*srcH;
    byte *pDestHalfV	= pDstY+iDstStride*srcH*3/2;
    byte *pDestEighthU1	= pDstU+iDstStrideUV*srcH/2;
    byte *pDestEighthV1	= pDstU+iDstStrideUV*srcH*3/4;
    byte *pDestEighthU2	= pDstV+iDstStrideUV*srcH/2;
    byte *pDestEighthV2	= pDstV+iDstStrideUV*srcH*3/4;

    byte* pY = pSrcInp;
    byte* pU = pSrcInp + srcH * srcW;
    byte* pV = pSrcInp + srcH * srcW * 2;

    //Copy over Y for main channel
    memcpy(pDestY420,pSrcInp,srcH*srcW);

    //Frame packing for Chroma channels
    for ( y = 0; y < srcH; ++ y ) 
    {
        for ( x = 0; x < srcW; ++ x ) 
        {
            if (y & 1) {
                *(pDestHalfU+x) = *pU;
                *(pDestHalfV+x) = *pV;
            }
            else if (1==(x&1 ) && 0==(y&2)) {
                *(pDestEighthU1+x/2) = *pU;
                *(pDestEighthV1+x/2) = *pV;
            }
            else if (1==(x&1 ) && 2==(y&2)){
                *(pDestEighthU2+x/2) = *pU;
                *(pDestEighthV2+x/2) = *pV;
            }
            else { // ( !(x & 1) && !(y & 1)) 
                if(bAntiAlias) {
                    //Simple averaging used as anti-alias filter for now. Modify if needed in future
                    *(pDestU420+x/2) = (*pU + *(pU+1) + *(pU+iSrcStrideUV) + *(pU+iSrcStrideUV+1) + 2) >> 2;
                    *(pDestV420+x/2) = (*pV + *(pV+1) + *(pV+iSrcStrideUV) + *(pV+iSrcStrideUV+1) + 2) >> 2;
                }
                else{
                    *(pDestU420+x/2) = *pU; 
                    *(pDestV420+x/2) = *pV;
                }
            }

            pU++;
            pV++;
        }

        if ( y & 1 ) {
            pDestHalfU += iDstStride;
            pDestHalfV += iDstStride;
        } 
        else {
            pDestU420 += iDstStrideUV;
            pDestV420 += iDstStrideUV;
            if ( 0 == ( y & 2 ) ) {
                pDestEighthU1 += iDstStrideUV;
                pDestEighthV1 += iDstStrideUV;
            }
            else {
                pDestEighthU2 += iDstStrideUV;
                pDestEighthV2 += iDstStrideUV;
            }
        }
    }

}

void FramePackBandSeparation(byte *pSrc, byte *pDstY, byte *pDstU, byte *pDstV, int srcH, int srcW, 
                         int iSrcStrideUV, int iDstStride,int iDstStrideUV, double dAlpha, double dBeta)
{
    int x,y;
    byte *pSrcInp       = pSrc;
    byte *pDestY420		= pDstY;
    byte *pDestU420		= pDstU;
    byte *pDestV420		= pDstV;
    byte *pDestHalfU    = pDstY+iDstStride*srcH;
    byte *pDestHalfV	= pDstY+iDstStride*srcH*3/2;
    byte *pDestEighthU1	= pDstU+iDstStrideUV*srcH/2;
    byte *pDestEighthV1	= pDstU+iDstStrideUV*srcH*3/4;
    byte *pDestEighthU2	= pDstV+iDstStrideUV*srcH/2;
    byte *pDestEighthV2	= pDstV+iDstStrideUV*srcH*3/4;
    int **ppCU, **ppCV, **ppDU, **ppDV;
    ppCU = new int*[srcH/ 2];
    ppDU = new int*[srcH / 2];
    ppCV = new int*[srcH / 2];
    ppDV = new int*[srcH / 2];
    for ( y = 0; y < srcH / 2; ++ y )
    {
        ppCU[y] = new int[srcW];
        ppCV[y] = new int[srcW];
        ppDU[y] = new int[srcW];
        ppDV[y] = new int[srcW];
    }

    byte* pY = pSrcInp;
    byte* pU = pSrcInp + srcH * srcW;
    byte* pV = pSrcInp + srcH * srcW * 2;

    //Copy over Y for main channel
    memcpy(pDestY420,pSrcInp,srcH*srcW);

    //Pre-computing intermediate samples
    for ( y = 0; y < srcH; ++ y ) 
    {
        for ( x = 0; x < srcW; ++ x ) 
        {
            if ( 0 == (y & 1) ) {
                ppCU[y>>1][x] = (*pU + *(pU+iSrcStrideUV));
                ppCV[y>>1][x] = (*pV + *(pV+iSrcStrideUV));
                ppDU[y>>1][x] = (*pU - *(pU+iSrcStrideUV));
                ppDV[y>>1][x] = (*pV - *(pV+iSrcStrideUV));
            }
                
            pU++;
            pV++;
        }
    }

    //Band Separation
    for ( y = 0; y < srcH/2; ++ y ) 
    {
        for ( x = 0; x < srcW; ++ x ) 
        {
            int diffU,diffV;
            if(0 == (x&1)) {
                *(pDestU420+x/2) = (ppCU[y][x] + ppCU[y][x+1] + 2) >> 2;
                *(pDestV420+x/2) = (ppCV[y][x] + ppCV[y][x+1] + 2) >> 2;
                    
                diffU = (ppCU[y][x] - ppCU[y][x+1]);
                diffV = (ppCV[y][x] - ppCV[y][x+1]);

                if (0==(y & 1) ) {
                    *(pDestEighthU1+x/2) = ClipToByte( ((diffU+2)>>2) + 128);
                    *(pDestEighthV1+x/2) = ClipToByte( ((diffV+2)>>2) + 128);
                }
                else {
                    *(pDestEighthU2+x/2) = ClipToByte( ((diffU+2)>>2) + 128);
                    *(pDestEighthV2+x/2) = ClipToByte( ((diffV+2)>>2) + 128);
                }
            }

            *(pDestHalfU+x) = ClipToByte(((ppDU[y][x]+1)>>1) + 128);
            *(pDestHalfV+x) = ClipToByte(((ppDV[y][x]+1)>>1) + 128);
        }
        pDestHalfU  += iDstStride;
        pDestHalfV  += iDstStride;
        pDestU420   += iDstStrideUV;
        pDestV420   += iDstStrideUV;
        if ( 0== ( y & 1 ) ) {
            pDestEighthU1 += iDstStrideUV;
            pDestEighthV1 += iDstStrideUV;
        }
        else {
            pDestEighthU2 += iDstStrideUV;
            pDestEighthV2 += iDstStrideUV;
        }
    }
    for ( y = 0; y < srcH / 2; ++ y ) {
        delete [] ppCU[y];
        delete [] ppCV[y];
        delete [] ppDU[y];
        delete [] ppDV[y];
    }

    delete ppCU;
    delete ppCV;
    delete ppDU;
    delete ppDV;

}

void FramePackLiftedBandSeparation(byte *pSrc, byte *pDstY, byte *pDstU, byte *pDstV, int srcH, int srcW, 
                         int iSrcStrideUV, int iDstStride,int iDstStrideUV, double dAlpha, double dBeta)
{
    int x,y;
    byte *pSrcInp       = pSrc;
    byte *pDestY420		= pDstY;
    byte *pDestU420		= pDstU;
    byte *pDestV420		= pDstV;
    byte *pDestHalfU    = pDstY+iDstStride*srcH;
    byte *pDestHalfV	= pDstY+iDstStride*srcH*3/2;
    byte *pDestEighthU1	= pDstU+iDstStrideUV*srcH/2;
    byte *pDestEighthV1	= pDstU+iDstStrideUV*srcH*3/4;
    byte *pDestEighthU2	= pDstV+iDstStrideUV*srcH/2;
    byte *pDestEighthV2	= pDstV+iDstStrideUV*srcH*3/4;

    byte* pY = pSrcInp;
    byte* pU = pSrcInp + srcH * srcW;
    byte* pV = pSrcInp + srcH * srcW * 2;
    
    //Copy over Y for main channel
    memcpy(pDestY420,pSrcInp,srcH*srcW);

    //Band Separation
    for ( y = 0; y < srcH; y+=2,pU+=2*iSrcStrideUV,pV+=2*iSrcStrideUV) 
    {
        for ( x = 0; x < srcW; x+=2) 
        {
            byte A,B,C,D;
            int t0,t1,t2;
            byte L0,L1;
            byte LLU,LHU,HU[2],LLV,LHV,HV[2];

            //U
            A = *(pU+x); 
            B = *(pU+x+1); 
            C = *(pU+iSrcStrideUV+x); 
            D = *(pU+iSrcStrideUV+x+1);

            t0 = A - C;
            t1 = B - D;
            L0 = (byte)((int)C + (t0>>1));
            L1 = (byte)((int)D + (t1>>1));
            t2 = L0 - L1;
            LLU = L1 + (t2>>1);
            LHU = ClipToByte(t2 +128);
            HU[0] = ClipToByte(t0 + 128);
            HU[1] = ClipToByte(t1 + 128);
            
            //V
            A = *(pV+x); 
            B = *(pV+x+1); 
            C = *(pV+iSrcStrideUV+x); 
            D = *(pV+iSrcStrideUV+x+1);

            t0 = A - C;
            t1 = B - D;
            L0 = (byte)((int)C + (t0>>1));
            L1 = (byte)((int)D + (t1>>1));
            t2 = L0 - L1;
            LLV = L1 + (t2>>1);
            LHV = ClipToByte(t2 +128);
            HV[0] = ClipToByte(t0 +128);
            HV[1] = ClipToByte(t1 +128);

            *(pDestU420+x/2) = LLU;
            *(pDestV420+x/2) = LLV;

            *(pDestHalfU+x)     = HU[0];
            *(pDestHalfU+x+1)   = HU[1];
            *(pDestHalfV+x)     = HV[0];
            *(pDestHalfV+x+1)   = HV[1];

            if (0==((y>>1) & 1) ) {
                *(pDestEighthU1+x/2) = LHU;
                *(pDestEighthV1+x/2) = LHV;
            }
            else {
                *(pDestEighthU2+x/2) = LHU;
                *(pDestEighthV2+x/2) = LHV;
            }
        }
        pDestHalfU  += iDstStride;
        pDestHalfV  += iDstStride;
        pDestU420   += iDstStrideUV;
        pDestV420   += iDstStrideUV;
        if ( 0== ((y>>1) & 1) ) {
            pDestEighthU1 += iDstStrideUV;
            pDestEighthV1 += iDstStrideUV;
        }
        else {
            pDestEighthU2 += iDstStrideUV;
            pDestEighthV2 += iDstStrideUV;
        }
    }

}

//------------------------------------------------------------------
// FRAME PACKING MODULES
//------------------------------------------------------------------
//NOTE: Only works with YUV/RGB that have even number of samples every row
void FramePacking( char *InputName, char *OutName, int srcH, int srcW, double dAlpha, double dBeta, FileType eFileType, bool bAntiAlias, FramePackingMode framePackingMode )
{
    FILE *pSrcFile = fopen( InputName, "rb" );
    FILE *pDstFile = fopen( OutName, "wb" );
    int iFactorFileType;
    if ( RGB32 == eFileType )
        iFactorFileType = 4;
    else if ( YUV444 == eFileType )
        iFactorFileType = 3;
    else {
        printf("\n Unsupported mode");
        return;
    }

    byte *pSrc = new byte[srcH * srcW * iFactorFileType];
    byte *pDstY = new byte[srcH * srcW * 2];
    byte *pDstU = new byte[srcH * srcW / 2];
    byte *pDstV = new byte[srcH * srcW / 2];
    byte *pSrcYuv=NULL;
    int iCnt = 0;
    printf( "Finished:      " );

    while ( !feof( pSrcFile ) )
    {
        byte *pSrcInp = pSrc;
        fread( pSrc, sizeof(byte), srcH * srcW * iFactorFileType, pSrcFile );
        if ( feof(pSrcFile)) {
            break;
        }
        if ( RGB32 == eFileType) {
            pSrcYuv = new byte[srcH * srcW * 3];
            ColorConversionRGBtoYUV444( pSrc,pSrcYuv, srcH, srcW,srcW);
            pSrcInp = pSrcYuv;
        }

        if(framePackingMode == DirectMode)
            FramePackDirectMode(pSrcInp, pDstY,pDstU, pDstV,srcH, srcW,srcW,srcW,(srcW/2),dAlpha, dBeta,bAntiAlias);
        else if(framePackingMode == BandSeparationMode)
            FramePackBandSeparation(pSrcInp, pDstY,pDstU, pDstV,srcH, srcW,srcW,srcW,(srcW/2),dAlpha, dBeta);
        else if(framePackingMode == LiftedBandSeparationMode)
            FramePackLiftedBandSeparation(pSrcInp, pDstY,pDstU, pDstV,srcH, srcW,srcW,srcW,(srcW/2),dAlpha, dBeta);
        else
            printf("Not supported");

        fwrite( pDstY, sizeof(byte), srcH * srcW * 2, pDstFile );
        fwrite( pDstU, sizeof(byte), srcH * srcW / 2, pDstFile );
        fwrite( pDstV, sizeof(byte), srcH * srcW / 2, pDstFile );
        iCnt ++;
        printf( "\b\b\b\b\b%04d.", iCnt );
    }
    printf( "\n" );
    if ( pSrcYuv != NULL )
        delete[] pSrcYuv;

    delete[] pSrc;
    delete[] pDstY;
    delete[] pDstU;
    delete[] pDstV;
    fclose( pSrcFile );
    fclose( pDstFile );
}


void FrameUnpackDirectMode(byte *pSrcYUV420, byte *pDstY,byte *pDstU,byte *pDstV, int srcW, int srcH, int dwStrideSrcY, int dwStrideSrcUV, double dAlpha, double dBeta)
{
    int x,y;
    byte *pMY               = pDstY;
    byte *pMU               = pDstU;
    byte *pMV               = pDstV;
    byte *pSrc420Y          = pSrcYUV420;
    byte *pSrc420U          = pSrc420Y + srcH * srcW * 2;
    byte *pSrc420V          = pSrc420U + srcH * srcW / 2;
    byte *pSrcHalfU         = pSrc420Y + srcH * srcW; 
    byte *pSrcHalfV         = pSrc420Y + srcH * srcW * 3 / 2;
    byte *pSrcEighthU1	    = pSrc420U + srcH * dwStrideSrcUV / 2;
    byte *pSrcEighthV1	    = pSrc420U + srcH * dwStrideSrcUV * 3 / 4;
    byte *pSrcEighthU2	    = pSrc420V + srcH * dwStrideSrcUV / 2;
    byte *pSrcEighthV2	    = pSrc420V + srcH * dwStrideSrcUV * 3 / 4;

    //Copy over Y for main channel
    memcpy(pMY,pSrcYUV420,srcH*srcW);

    double tmp;
    for ( y = 0; y < srcH / 2 ; ++ y ) 
    {
        for ( x = 0; x < srcW / 2; ++ x ) 
        {
            int A,B,C,D;
            //U
            *(pMU + 2*y*dwStrideSrcY + 2*x)         = A = *(pSrc420U + y * dwStrideSrcUV + x);
            *(pMU + (2*y+1)*dwStrideSrcY + 2*x )    = B = (*( pSrcHalfU + y * dwStrideSrcY + 2*x));
            *(pMU + (2*y+1)*dwStrideSrcY + 2*x + 1 )= D = (*( pSrcHalfU + y * dwStrideSrcY + 2*x + 1));
            if(y%2 == 0)
                *(pMU + 2*y*dwStrideSrcY + 2*x + 1) = C = (*(pSrcEighthU1 + dwStrideSrcUV * (y>>1) + x ) );
            else
                *(pMU + 2*y*dwStrideSrcY + 2*x + 1) = C = (*(pSrcEighthU2 + dwStrideSrcUV * (y>>1) + x ) );

            // post filtering (NOTE that is alpha == beta == gamma == 0 implies no filtering)
            tmp = (( ( 1 + dAlpha * 2 + dBeta ) * A - dAlpha * (B + C) - dBeta * D ) + 0.5 );
            tmp = (tmp > 255.0 ? 255 : (tmp < 0.0 ? 0 : tmp));
            *(pMU + 2*y*dwStrideSrcY + 2*x) = (byte)tmp;

            //V
            *(pMV + 2*y*dwStrideSrcY + 2*x)         = A = *(pSrc420V + y * dwStrideSrcUV + x);
            *(pMV + (2*y+1)*dwStrideSrcY + 2*x )    = B = (*( pSrcHalfV + y * dwStrideSrcY + 2*x));
            *(pMV + (2*y+1)*dwStrideSrcY + 2*x + 1 )= D = (*( pSrcHalfV + y * dwStrideSrcY + 2*x + 1));
            if(y%2 == 0)
                *(pMV + 2*y*dwStrideSrcY + 2*x + 1) = C = (*(pSrcEighthV1 + dwStrideSrcUV * (y>>1) + x ) );
            else
                *(pMV + 2*y*dwStrideSrcY + 2*x + 1) = C = (*(pSrcEighthV2 + dwStrideSrcUV * (y>>1) + x ) );

            // post filtering (NOTE that is alpha == beta == gamma == 0 implies no filtering)
            tmp = (( ( 1 + dAlpha * 2 + dBeta ) * A - dAlpha * (B + C) - dBeta * D ) + 0.5 );
            tmp = (tmp > 255.0 ? 255 : (tmp < 0.0 ? 0 : tmp));
            *(pMV + 2*y*dwStrideSrcY + 2*x) = (byte)tmp;
        }
    }    
}

void FrameUnpackBandSeparation(byte *pSrcYUV420, byte *pDstY,byte *pDstU,byte *pDstV, int srcW, int srcH, int dwStrideSrcY, int dwStrideSrcUV)
{
    int x,y;

    byte *pMY               = pDstY;
    byte *pMU               = pDstU;
    byte *pMV               = pDstV;
    byte *pSrc420Y          = pSrcYUV420;
    byte *pSrc420U          = pSrc420Y + srcH * srcW * 2;
    byte *pSrc420V          = pSrc420U + srcH * srcW / 2;
    byte *pSrcHalfU         = pSrc420Y + srcH * srcW; 
    byte *pSrcHalfV         = pSrc420Y + srcH * srcW * 3 / 2;
    byte *pSrcEighthU1	    = pSrc420U + srcH * dwStrideSrcUV / 2;
    byte *pSrcEighthV1	    = pSrc420U + srcH * dwStrideSrcUV * 3 / 4;
    byte *pSrcEighthU2	    = pSrc420V + srcH * dwStrideSrcUV / 2;
    byte *pSrcEighthV2	    = pSrc420V + srcH * dwStrideSrcUV * 3 / 4;

    int **ppCU, **ppCV, **ppDU, **ppDV;
    ppCU = new int*[srcH / 2];
    ppDU = new int*[srcH / 2];
    ppCV = new int*[srcH / 2];
    ppDV = new int*[srcH / 2];
    for ( y = 0; y < srcH / 2; ++ y ) {
        ppCU[y] = new int[srcW];
        ppCV[y] = new int[srcW];
        ppDU[y] = new int[srcW];
        ppDV[y] = new int[srcW];
    }

    //Copy over Y for main channel
    memcpy(pMY,pSrcYUV420,srcH*srcW);

    for ( x = 0; x < (srcW >> 1); ++ x ) {
        for ( y = 0; y < (srcH >> 1); ++ y ) {
            int LLCb,LCb[2],HCb[2],HHCb;
            byte A,B,C,D;

            //U
            LLCb = *(pSrc420U + y * dwStrideSrcUV + x);
            HCb[0] = ((int)(*( pSrcHalfU + y * dwStrideSrcY + 2*x))) - 128;
            HCb[1] = ((int)(*( pSrcHalfU + y * dwStrideSrcY + 2*x + 1))) - 128;
            if(y%2 == 0)
                HHCb = ( (int)(*(pSrcEighthU1 + dwStrideSrcUV * (y>>1) + x ) ) ) - 128;
            else
                HHCb = ( (int)(*(pSrcEighthU2 + dwStrideSrcUV * (y>>1) + x ) ) ) - 128;
            LCb[0] = LLCb + HHCb;
            LCb[1] = LLCb - HHCb;
            *(pMU + 2*y*dwStrideSrcY + 2*x)         = A = ClipToByte(LCb[0] + HCb[0]);
            *(pMU + (2*y+1)*dwStrideSrcY + 2*x )    = B = ClipToByte(LCb[0] - HCb[0]);
            *(pMU + 2*y*dwStrideSrcY + 2*x + 1 )    = C = ClipToByte(LCb[1] + HCb[1]);
            *(pMU + (2*y+1)*dwStrideSrcY + 2*x + 1 )= D = ClipToByte(LCb[1] - HCb[1]);

            // No post filtering for band-separated case

            //V
            LLCb = *(pSrc420V + y * dwStrideSrcUV + x);
            HCb[0] = ((int)(*( pSrcHalfV + y * dwStrideSrcY + 2*x))) - 128;
            HCb[1] = ((int)(*( pSrcHalfV + y * dwStrideSrcY + 2*x + 1))) - 128;
            if(y%2 == 0)
                HHCb = ( (int)(*(pSrcEighthV1 + dwStrideSrcUV * (y>>1) + x))) - 128;
            else
                HHCb = ( (int)(*(pSrcEighthV2 + dwStrideSrcUV * (y>>1) + x))) - 128;
            LCb[0] = LLCb + HHCb;
            LCb[1] = LLCb - HHCb;
            *(pMV + 2*y*dwStrideSrcY + 2*x)         = A = ClipToByte(LCb[0] + HCb[0]);
            *(pMV + (2*y+1)*dwStrideSrcY + 2*x )    = B = ClipToByte(LCb[0] - HCb[0]);
            *(pMV + 2*y*dwStrideSrcY + 2*x + 1 )    = C = ClipToByte(LCb[1] + HCb[1]);
            *(pMV + (2*y+1)*dwStrideSrcY + 2*x + 1 )= D = ClipToByte(LCb[1] - HCb[1]);

            // No post filtering for band-separated case
        }
    }

    for ( y = 0; y < srcH / 2; ++ y ) {
        delete [] ppCU[y];
        delete [] ppCV[y];
        delete [] ppDU[y];
        delete [] ppDV[y];
    }

    delete ppCU;
    delete ppCV;
    delete ppDU;
    delete ppDV;
}


void FrameUnpackLiftedBandSeparation(byte *pSrcYUV420, byte *pDstY,byte *pDstU,byte *pDstV, int srcW, int srcH, int dwStrideSrcY, int dwStrideSrcUV)
{
    int x,y;

    byte *pMY               = pDstY;
    byte *pMU               = pDstU;
    byte *pMV               = pDstV;
    byte *pSrc420Y          = pSrcYUV420;
    byte *pSrc420U          = pSrc420Y + srcH * srcW * 2;
    byte *pSrc420V          = pSrc420U + srcH * srcW / 2;
    byte *pSrcHalfU         = pSrc420Y + srcH * srcW; 
    byte *pSrcHalfV         = pSrc420Y + srcH * srcW * 3 / 2;
    byte *pSrcEighthU1	    = pSrc420U + srcH * dwStrideSrcUV / 2;
    byte *pSrcEighthV1	    = pSrc420U + srcH * dwStrideSrcUV * 3 / 4;
    byte *pSrcEighthU2	    = pSrc420V + srcH * dwStrideSrcUV / 2;
    byte *pSrcEighthV2	    = pSrc420V + srcH * dwStrideSrcUV * 3 / 4;

    //Copy over Y for main channel
    memcpy(pMY,pSrcYUV420,srcH*srcW);

    for ( x = 0; x < (srcW >> 1); ++ x ) {
        for ( y = 0; y < (srcH >> 1); ++ y ) {
            int LLCb,LCb[2],HCb[2],HHCb;
            byte A,B,C,D;

            //U
            LLCb = *(pSrc420U + y * dwStrideSrcUV + x);
            HCb[0] = ((int)(*( pSrcHalfU + y * dwStrideSrcY + 2*x))) - 128;
            HCb[1] = ((int)(*( pSrcHalfU + y * dwStrideSrcY + 2*x + 1))) - 128;
            if(y%2 == 0)
                HHCb = ( (int)(*(pSrcEighthU1 + dwStrideSrcUV * (y>>1) + x ) ) ) - 128;
            else
                HHCb = ( (int)(*(pSrcEighthU2 + dwStrideSrcUV * (y>>1) + x ) ) ) - 128;

            LCb[1] = ClipToByte(LLCb - (HHCb>>1));
            LCb[0] = ClipToByte(LCb[1] + HHCb);

            *(pMU + (2*y+1)*dwStrideSrcY + 2*x )    = C = ClipToByte(LCb[0] - (HCb[0]>>1));
            *(pMU + 2*y*dwStrideSrcY + 2*x)             = A = ClipToByte(C + HCb[0]);
            *(pMU + (2*y+1)*dwStrideSrcY + 2*x + 1 )    = D = ClipToByte(LCb[1] - (HCb[1]>>1));
            *(pMU + (2*y)*dwStrideSrcY + 2*x + 1)       = B = ClipToByte(D + HCb[1]);

            // No post filtering for lifted case

            //V
            LLCb = *(pSrc420V + y * dwStrideSrcUV + x);
            HCb[0] = ((int)(*( pSrcHalfV + y * dwStrideSrcY + 2*x))) - 128;
            HCb[1] = ((int)(*( pSrcHalfV + y * dwStrideSrcY + 2*x + 1))) - 128;
            if(y%2 == 0)
                HHCb = ( (int)(*(pSrcEighthV1 + dwStrideSrcUV * (y>>1) + x ) ) ) - 128;
            else
                HHCb = ( (int)(*(pSrcEighthV2 + dwStrideSrcUV * (y>>1) + x ) ) ) - 128;

            LCb[1] = ClipToByte(LLCb - (HHCb>>1));
            LCb[0] = ClipToByte(LCb[1] + HHCb);

            *(pMV + (2*y+1)*dwStrideSrcY + 2*x )    = C = ClipToByte(LCb[0] - (HCb[0]>>1));
            *(pMV + 2*y*dwStrideSrcY + 2*x)             = A = ClipToByte(C + HCb[0]);
            *(pMV + (2*y+1)*dwStrideSrcY + 2*x + 1 )    = D = ClipToByte(LCb[1] - (HCb[1]>>1));
            *(pMV + (2*y)*dwStrideSrcY + 2*x + 1)       = B = ClipToByte(D + HCb[1]);

            // No post filtering for lifted case        
        }
    }

}

void FrameUnpacking( char *InputName, char *OutName, int srcH, int srcW, double dAlpha, double dBeta, FileType eFileType, FramePackingMode framePackingMode)
{
    FILE *pSrcFile = fopen( InputName, "rb" );
    FILE *pDstFile = fopen( OutName, "wb" );
    int iFactorFileType;
    byte* pDstY, *pDstU, *pDstV, *pDstBGR=NULL;
    pDstY = new byte[srcH * srcW];
    pDstU = new byte[srcH * srcW];
    pDstV = new byte[srcH * srcW];
    if ( RGB32 == eFileType )
    {
        iFactorFileType = 4;
        pDstBGR = new byte[srcH * srcW * 4];
    }
    else if ( YUV444 == eFileType )
    {
        iFactorFileType = 3;
    }
    else {
        printf("\n Unsupported mode");
        return;
    }

    byte *pSrcYUV420 = new byte[srcH * srcW * 3];
    int iCnt = 0;
    printf( "Finished:      " );
    while ( !feof( pSrcFile ) ) 
    {
        fread( pSrcYUV420, sizeof(byte), srcH * srcW * 3, pSrcFile );
        if ( feof( pSrcFile ) )
        {
            break;
        }

        if(framePackingMode == DirectMode)
            FrameUnpackDirectMode(pSrcYUV420, pDstY,pDstU,pDstV, srcW,srcH,srcW,(srcW/2),dAlpha,dBeta);
        else if(framePackingMode == BandSeparationMode)
            FrameUnpackBandSeparation(pSrcYUV420, pDstY,pDstU,pDstV, srcW,srcH,srcW,(srcW/2));
        else if(framePackingMode == LiftedBandSeparationMode)
            FrameUnpackLiftedBandSeparation(pSrcYUV420, pDstY,pDstU,pDstV, srcW,srcH,srcW,(srcW/2));
        else
            printf("Other modes not supported yet");

        if ( RGB32 == eFileType )
        {
            ColorConversionYUV444toRGB32(pDstY,pDstU,pDstV,pDstBGR,srcW,srcH,srcW);
            fwrite( pDstBGR, sizeof(byte), srcH * srcW * 4, pDstFile );
        }
        else
        {
            fwrite( pDstY, sizeof(byte), srcH * srcW, pDstFile );
            fwrite( pDstU, sizeof(byte), srcH * srcW, pDstFile );
            fwrite( pDstV, sizeof(byte), srcH * srcW, pDstFile );
        }
        iCnt ++;
        printf( "\b\b\b\b\b%04d.", iCnt );
    }
    printf( "\n" );
    delete [] pSrcYUV420;
    if ( pDstBGR != NULL)
    {
        delete [] pDstBGR;
    }
    delete [] pDstY;
    delete [] pDstU;
    delete [] pDstV;
    fclose( pSrcFile );
    fclose( pDstFile );
}


void EvaluateFramePacking( char *InputName, int srcH, int srcW, double dAlpha, double dBeta, FileType eFileType, FramePackingMode framePackingMode)
{
    int iFactorFileType;
    int cnt=0;
    double psnrY,psnrU,psnrV,totPSNRY=0,totPSNRU=0,totPSNRV=0;
    FILE *pSrcFile = fopen( InputName, "rb" );
    byte *pSrc, *pSrcYuv, *pFinal;
    byte *pSrcInp;
    byte *p420YUV = new byte[srcH * srcW * 3];
    byte *pfinDst = new byte[srcH * srcW * 3];

    byte *pY420 = p420YUV;
    byte *pU420 = pY420 + (srcH * srcW * 2);
    byte *pV420 = pU420 + srcH * srcW / 2;

    byte *pfinY = pfinDst;
    byte *pfinU = pfinY + (srcH * srcW);
    byte *pfinV = pfinU + (srcH * srcW);


    if ( RGB32 == eFileType ) {
        iFactorFileType = 4;
        pSrc = new byte[srcH * srcW * 4];
        pFinal = new byte[srcH * srcW * 4];
    }
    else {
        iFactorFileType = 3;
        pSrc = new byte[srcH * srcW * 3];
        pFinal = new byte[srcH * srcW * 3];
    }
    int srcStride = srcW;
    int srcStrideUV = srcW/ 2;
    int dstStrideY  = srcW;
    int dstStrideUV = srcW / 2;

    while ( !feof( pSrcFile ) )
    {
        pSrcInp = pSrc;
        fread( pSrc, sizeof(byte), srcH * srcW * iFactorFileType, pSrcFile );
        if ( feof(pSrcFile)) {
            break;
        }
        if ( RGB32 == eFileType) {
            pSrcYuv = new byte[srcH * srcW * 3];
            ColorConversionRGBtoYUV444( pSrc,pSrcYuv, srcH, srcW,srcW);
            pSrcInp = pSrcYuv;
        }
        else
            pSrcInp = pSrc;

        if(framePackingMode == DirectMode) {
            FramePackDirectMode(pSrcInp, pY420,pU420, pV420,srcH, srcW,srcStrideUV,dstStrideY,dstStrideUV,dAlpha, dBeta,PERFORM_ANTI_ALIASING);
            FrameUnpackDirectMode(p420YUV, pfinY,pfinU, pfinV,srcW, srcH,dstStrideY,dstStrideUV,dAlpha, dBeta);
        }
        else if(framePackingMode == BandSeparationMode) {
            FramePackBandSeparation(pSrcInp, pY420,pU420, pV420,srcH, srcW,srcStrideUV,dstStrideY,dstStrideUV,dAlpha, dBeta);
            FrameUnpackBandSeparation(p420YUV, pfinY,pfinU, pfinV,srcW, srcH,dstStrideY,dstStrideUV);
        }
        else if(framePackingMode == LiftedBandSeparationMode) {
            FramePackLiftedBandSeparation(pSrcInp, pY420,pU420, pV420,srcH, srcW,srcStrideUV,dstStrideY,dstStrideUV,dAlpha, dBeta);
            FrameUnpackLiftedBandSeparation(p420YUV, pfinY,pfinU, pfinV,srcW, srcH,dstStrideY,dstStrideUV);
        }
        else
            printf("Not supported");

        ComputePSNRforYUV(pSrcInp,pfinDst,srcW,srcH,YUV_444,&psnrY,&psnrU,&psnrV);

        totPSNRY += psnrY;
        totPSNRU += psnrU;
        totPSNRV += psnrV;

        printf("\n %d, PSNR_U: %f, PSNR_V: %f",cnt++,psnrU,psnrV);
    }
    printf( "\n" );
    printf("\n Average PSNR_Y\n PSNR_U: %f, PSNR_V: %f, PSNR_UV: %f",totPSNRU/cnt,totPSNRV/cnt, (totPSNRU+totPSNRV)/(cnt*2));
    if ( RGB32 == eFileType )
        delete[] pSrcYuv;

    delete[] pSrc;
    delete[] p420YUV;
    delete[] pfinDst;
    fclose( pSrcFile );
}


void ComputePSNR( char *InputName, char *OutName, int srcW, int srcH, FileType eFileType)
{
    int cnt=0;
    double psnrY,psnrU,psnrV,totPSNRY=0,totPSNRU=0,totPSNRV=0;
    FILE *pSrcFile = fopen( InputName, "rb" );
    FILE *pDstFile = fopen( OutName, "rb" );
    int iFactorFileType;
    int frameSize;
    if ( RGB32 == eFileType ){
        printf("\n ONly YUV psnr supported as of now");
        return;
    }
    else if ( YUV444 == eFileType ){
        iFactorFileType = 3;
        frameSize= (srcH * srcW * 3);
    }
    else{
        frameSize= (srcH * srcW * 3)/2;
    }

    byte *pSrc = new byte[srcH * srcW * iFactorFileType];
    byte *pDst = new byte[srcH * srcW * iFactorFileType];
    int iCnt = 0;

    while ( !feof( pSrcFile ) && !feof( pDstFile ))
    {
        fread( pSrc, sizeof(byte), frameSize, pSrcFile );
        fread( pDst, sizeof(byte), frameSize, pDstFile );
        if ( feof(pSrcFile) || feof(pDstFile)) {
            break;
        }
        if(YUV444 == eFileType )
            ComputePSNRforYUV(pSrc,pDst,srcW,srcH,YUV_444,&psnrY,&psnrU,&psnrV);
        else
            ComputePSNRforYUV(pSrc,pDst,srcW,srcH,YUV_420,&psnrY,&psnrU,&psnrV);
        
        totPSNRY += psnrY;
        totPSNRU += psnrU;
        totPSNRV += psnrV;

#ifdef DISPLAY_ALL_FRAMES_PSNR
        printf("\n %d, PSNR_Y: %f, PSNR_U: %f, PSNR_V: %f",cnt,psnrY, psnrU,psnrV);
#endif
        cnt++;
    }

    printf("\n Average PSNR: PSNR_Y: %f, PSNR_U: %f, PSNR_V: %f, PSNR_UV: %f",totPSNRY/cnt,totPSNRU/cnt,totPSNRV/cnt, (totPSNRU+totPSNRV)/(cnt*2));

    delete[] pSrc;
    delete[] pDst;
    fclose( pSrcFile );
    fclose( pDstFile );

}

